// This program interfaces keypad,LCD, Power LED and two peristaltic pump.
//The samples are pumped in the batches of 15ml and exposure
//key 1: mode change; key 2:Increment;Key 3: Decrement Key A:Start/stop
// List of modes: volume,intensity and exposure time
#include <avr/interrupt.h>
#include <avr/io.h> 
#include <Keypad.h>
#include <EEPROM.h>
//LCD Constants 
#define PIN_RESET 7
#define PIN_SCE   8
#define PIN_DC    9
#define PIN_SDIN  10
#define PIN_SCLK  12
#define LCD_C     LOW
#define LCD_D     HIGH
#define LCD_X     84
#define LCD_Y     48
//Keypad constants
const byte ROWS = 2; //two rows
const byte COLS = 4; //four columns

//Constants need attention
byte rowPins[ROWS] = {A0,A1}; //connect to the row pin
byte colPins[COLS] = {A5,A4,A3,A2};//connect to the column pinouts of the keypad
int pump1_rate=53,pump2_rate=54;
int pwmpins[3]={3,6,11}; //3(pump1) and 11(pump2) HF PWM and 6 for LED
int pump_dir_pin[2]={4,5}; //pin 4 for pump1,pin 5 for pump2
int pump_enable_pin=2;


//define the cymbols on the buttons of the keypads
char hexaKeys[ROWS][COLS] = {  {'1','2','3','A'  },{'4','5','6','B'}};
int temp;
byte digit,state=0;
//initialize an instance of class NewKeypad
Keypad customKeypad = Keypad( makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS);
int user_data[4]={0,0,0,0};
int  user_data_max[4]={90,100,300,1},user_data_min[4]={0,0,10,0};
int batch_volume=15,activated_volume=0,batches,total_duration,total_elapsed;
int pwmvalues[3],eprom_address[4]={0,4,8,12};
int current_mode=0;
int j=0,duration_start[3]={0,0,0},duration_end[3]={30,30,30},elapsed_time=0,pump_dir_value[2]={0,0};
float rate1=1.2,rate2=1.2;



static const byte ASCII[][5] =
{
  {
    0x00, 0x00, 0x00, 0x00, 0x00  } // 20  
  ,{
    0x00, 0x00, 0x5f, 0x00, 0x00  } // 21 !
  ,{
    0x00, 0x07, 0x00, 0x07, 0x00  } // 22 "
  ,{
    0x14, 0x7f, 0x14, 0x7f, 0x14  } // 23 #
  ,{
    0x24, 0x2a, 0x7f, 0x2a, 0x12  } // 24 $
  ,{
    0x23, 0x13, 0x08, 0x64, 0x62  } // 25 %
  ,{
    0x36, 0x49, 0x55, 0x22, 0x50  } // 26 &
  ,{
    0x00, 0x05, 0x03, 0x00, 0x00  } // 27 '
  ,{
    0x00, 0x1c, 0x22, 0x41, 0x00  } // 28 (
  ,{
    0x00, 0x41, 0x22, 0x1c, 0x00  } // 29 )
  ,{
    0x14, 0x08, 0x3e, 0x08, 0x14  } // 2a *
  ,{
    0x08, 0x08, 0x3e, 0x08, 0x08  } // 2b +
  ,{
    0x00, 0x50, 0x30, 0x00, 0x00  } // 2c ,
  ,{
    0x08, 0x08, 0x08, 0x08, 0x08  } // 2d -
  ,{
    0x00, 0x60, 0x60, 0x00, 0x00  } // 2e .
  ,{
    0x20, 0x10, 0x08, 0x04, 0x02  } // 2f /
  ,{
    0x3e, 0x51, 0x49, 0x45, 0x3e  } // 30 0
  ,{
    0x00, 0x42, 0x7f, 0x40, 0x00  } // 31 1
  ,{
    0x42, 0x61, 0x51, 0x49, 0x46  } // 32 2
  ,{
    0x21, 0x41, 0x45, 0x4b, 0x31  } // 33 3
  ,{
    0x18, 0x14, 0x12, 0x7f, 0x10  } // 34 4
  ,{
    0x27, 0x45, 0x45, 0x45, 0x39  } // 35 5
  ,{
    0x3c, 0x4a, 0x49, 0x49, 0x30  } // 36 6
  ,{
    0x01, 0x71, 0x09, 0x05, 0x03  } // 37 7
  ,{
    0x36, 0x49, 0x49, 0x49, 0x36  } // 38 8
  ,{
    0x06, 0x49, 0x49, 0x29, 0x1e  } // 39 9
  ,{
    0x00, 0x36, 0x36, 0x00, 0x00  } // 3a :
  ,{
    0x00, 0x56, 0x36, 0x00, 0x00  } // 3b ;
  ,{
    0x08, 0x14, 0x22, 0x41, 0x00  } // 3c <
  ,{
    0x14, 0x14, 0x14, 0x14, 0x14  } // 3d =
  ,{
    0x00, 0x41, 0x22, 0x14, 0x08  } // 3e >
  ,{
    0x02, 0x01, 0x51, 0x09, 0x06  } // 3f ?
  ,{
    0x32, 0x49, 0x79, 0x41, 0x3e  } // 40 @
  ,{
    0x7e, 0x11, 0x11, 0x11, 0x7e  } // 41 A
  ,{
    0x7f, 0x49, 0x49, 0x49, 0x36  } // 42 B
  ,{
    0x3e, 0x41, 0x41, 0x41, 0x22  } // 43 C
  ,{
    0x7f, 0x41, 0x41, 0x22, 0x1c  } // 44 D
  ,{
    0x7f, 0x49, 0x49, 0x49, 0x41  } // 45 E
  ,{
    0x7f, 0x09, 0x09, 0x09, 0x01  } // 46 F
  ,{
    0x3e, 0x41, 0x49, 0x49, 0x7a  } // 47 G
  ,{
    0x7f, 0x08, 0x08, 0x08, 0x7f  } // 48 H
  ,{
    0x00, 0x41, 0x7f, 0x41, 0x00  } // 49 I
  ,{
    0x20, 0x40, 0x41, 0x3f, 0x01  } // 4a J
  ,{
    0x7f, 0x08, 0x14, 0x22, 0x41  } // 4b K
  ,{
    0x7f, 0x40, 0x40, 0x40, 0x40  } // 4c L
  ,{
    0x7f, 0x02, 0x0c, 0x02, 0x7f  } // 4d M
  ,{
    0x7f, 0x04, 0x08, 0x10, 0x7f  } // 4e N
  ,{
    0x3e, 0x41, 0x41, 0x41, 0x3e  } // 4f O
  ,{
    0x7f, 0x09, 0x09, 0x09, 0x06  } // 50 P
  ,{
    0x3e, 0x41, 0x51, 0x21, 0x5e  } // 51 Q
  ,{
    0x7f, 0x09, 0x19, 0x29, 0x46  } // 52 R
  ,{
    0x46, 0x49, 0x49, 0x49, 0x31  } // 53 S
  ,{
    0x01, 0x01, 0x7f, 0x01, 0x01  } // 54 T
  ,{
    0x3f, 0x40, 0x40, 0x40, 0x3f  } // 55 U
  ,{
    0x1f, 0x20, 0x40, 0x20, 0x1f  } // 56 V
  ,{
    0x3f, 0x40, 0x38, 0x40, 0x3f  } // 57 W
  ,{
    0x63, 0x14, 0x08, 0x14, 0x63  } // 58 X
  ,{
    0x07, 0x08, 0x70, 0x08, 0x07  } // 59 Y
  ,{
    0x61, 0x51, 0x49, 0x45, 0x43  } // 5a Z
  ,{
    0x00, 0x7f, 0x41, 0x41, 0x00  } // 5b [
  ,{
    0x02, 0x04, 0x08, 0x10, 0x20  } // 5c ¥
  ,{
    0x00, 0x41, 0x41, 0x7f, 0x00  } // 5d ]
  ,{
    0x04, 0x02, 0x01, 0x02, 0x04  } // 5e ^
  ,{
    0x40, 0x40, 0x40, 0x40, 0x40  } // 5f _
  ,{
    0x00, 0x01, 0x02, 0x04, 0x00  } // 60 `
  ,{
    0x20, 0x54, 0x54, 0x54, 0x78  } // 61 a
  ,{
    0x7f, 0x48, 0x44, 0x44, 0x38  } // 62 b
  ,{
    0x38, 0x44, 0x44, 0x44, 0x20  } // 63 c
  ,{
    0x38, 0x44, 0x44, 0x48, 0x7f  } // 64 d
  ,{
    0x38, 0x54, 0x54, 0x54, 0x18  } // 65 e
  ,{
    0x08, 0x7e, 0x09, 0x01, 0x02  } // 66 f
  ,{
    0x0c, 0x52, 0x52, 0x52, 0x3e  } // 67 g
  ,{
    0x7f, 0x08, 0x04, 0x04, 0x78  } // 68 h
  ,{
    0x00, 0x44, 0x7d, 0x40, 0x00  } // 69 i
  ,{
    0x20, 0x40, 0x44, 0x3d, 0x00  } // 6a j 
  ,{
    0x7f, 0x10, 0x28, 0x44, 0x00  } // 6b k
  ,{
    0x00, 0x41, 0x7f, 0x40, 0x00  } // 6c l
  ,{
    0x7c, 0x04, 0x18, 0x04, 0x78  } // 6d m
  ,{
    0x7c, 0x08, 0x04, 0x04, 0x78  } // 6e n
  ,{
    0x38, 0x44, 0x44, 0x44, 0x38  } // 6f o
  ,{
    0x7c, 0x14, 0x14, 0x14, 0x08  } // 70 p
  ,{
    0x08, 0x14, 0x14, 0x18, 0x7c  } // 71 q
  ,{
    0x7c, 0x08, 0x04, 0x04, 0x08  } // 72 r
  ,{
    0x48, 0x54, 0x54, 0x54, 0x20  } // 73 s
  ,{
    0x04, 0x3f, 0x44, 0x40, 0x20  } // 74 t
  ,{
    0x3c, 0x40, 0x40, 0x20, 0x7c  } // 75 u
  ,{
    0x1c, 0x20, 0x40, 0x20, 0x1c  } // 76 v
  ,{
    0x3c, 0x40, 0x30, 0x40, 0x3c  } // 77 w
  ,{
    0x44, 0x28, 0x10, 0x28, 0x44  } // 78 x
  ,{
    0x0c, 0x50, 0x50, 0x50, 0x3c  } // 79 y
  ,{
    0x44, 0x64, 0x54, 0x4c, 0x44  } // 7a z
  ,{
    0x00, 0x08, 0x36, 0x41, 0x00  } // 7b {
  ,{
    0x00, 0x00, 0x7f, 0x00, 0x00  } // 7c |
  ,{
    0x00, 0x41, 0x36, 0x08, 0x00  } // 7d }
  ,{
    0x10, 0x08, 0x08, 0x10, 0x08  } // 7e ←
  ,{
    0x78, 0x46, 0x41, 0x46, 0x78  } // 7f →
};

void LcdCharacter(char character)
{
  LcdWrite(LCD_D, 0x00);
  for (int index = 0; index < 5; index++)
  {
    LcdWrite(LCD_D, ASCII[character - 0x20][index]);
  }
  LcdWrite(LCD_D, 0x00);
}

void LcdClear(void)
{
  for (int index = 0; index < LCD_X * LCD_Y / 8; index++)
  {
    LcdWrite(LCD_D, 0x00);
  }
}

void LcdInitialise(void)
{
  pinMode(PIN_SCE, OUTPUT);
  pinMode(PIN_RESET, OUTPUT);
  pinMode(PIN_DC, OUTPUT);
  pinMode(PIN_SDIN, OUTPUT);
  pinMode(PIN_SCLK, OUTPUT);
  digitalWrite(PIN_RESET, LOW);
  digitalWrite(PIN_RESET, HIGH);
  LcdWrite(LCD_C, 0x21 );  // LCD Extended Commands.
  LcdWrite(LCD_C, 0xB1 );  // Set LCD Vop (Contrast). 
  LcdWrite(LCD_C, 0x04 );  // Set Temp coefficent. //0x04
  LcdWrite(LCD_C, 0x14 );  // LCD bias mode 1:48. //0x13
  LcdWrite(LCD_C, 0x20 );  // LCD Basic Commands
  LcdWrite(LCD_C, 0x0C );  // LCD in normal mode.
}

void LcdString(char *characters)
{
  while (*characters)
  {
    LcdCharacter(*characters++);
  }
}

void LcdWrite(byte dc, byte data)
{
  digitalWrite(PIN_DC, dc);
  digitalWrite(PIN_SCE, LOW);
  shiftOut(PIN_SDIN, PIN_SCLK, MSBFIRST, data);
  digitalWrite(PIN_SCE, HIGH);
}
void gotoXY(int x, int y)
{
  LcdWrite( 0, 0x80 | x);  // Column.
  LcdWrite( 0, 0x40 | y);  // Row.  

}
void LcdInt(int number) // writes 3 digit number in the LCD
{
  int temp_var=number;
  int count=0;
  byte digit[3]={
    0,0,0  };
  do
  {
    digit[count]=temp_var%10;
    temp_var=temp_var/10;
    count++;
  } 
  while(temp_var!=0);
  if(digit[2]!=0)
  {
    LcdWrite(LCD_D, 0x00);
    for (int index = 0; index < 5; index++)
    {
      LcdWrite(LCD_D, ASCII[digit[2]+0x10][index]); //writes individual charaters
    }
    LcdWrite(LCD_D, 0x00);
  }

  for(j=1;j>=0;j--)
  {
    LcdWrite(LCD_D, 0x00);
    for (int index = 0; index < 5; index++)
    {
      LcdWrite(LCD_D, ASCII[digit[j]+0x10][index]);
    }
    LcdWrite(LCD_D, 0x00);
  }  
}
void LcdUpdate()
{
  //Mode Name
  gotoXY(0,1);
  if(current_mode==0)
    LcdString("Mode:Volume");
  else if(current_mode==1)
    LcdString("Mode:INT   ");
  else if(current_mode==2)
    LcdString("Mode:Time  ");
  else if(current_mode==3)
    LcdString("Mode:Stir  ");


  //Values update.
  gotoXY(0,2);
  LcdString("Volume:");
  LcdInt(user_data[0]); 

  gotoXY(0,3);
  LcdString("LED:");
  LcdInt(user_data[1]);
  LcdString("/");
  LcdInt(user_data[2]);
  LcdString(" ");


  gotoXY(0,4);
  LcdString("Stir:");
  if(user_data[3]==0)
    LcdString("off  ");
  else if(user_data[3]==1)
    LcdString("on   ");

}

//KEYPAD functions
//for the key 1
void Modechange(void)
{
  if(current_mode<3)
    current_mode++;
  else if(current_mode>=3)
    current_mode=0;
  LcdUpdate();  
}

//for the key2
void Increase(void)
{
  if((user_data[current_mode]<user_data_max[current_mode])&&(user_data[current_mode]>=user_data_min[current_mode])&& current_mode==2)
    user_data[current_mode]+=10;
  else if(current_mode==3)
    user_data[current_mode]=1;       
  else if((user_data[current_mode]<user_data_max[current_mode])&&(user_data[current_mode]>=user_data_min[current_mode])) 
    user_data[current_mode]+=5;
  if(user_data[current_mode]>user_data_max[current_mode])
    user_data[current_mode]=user_data_max[current_mode];
  LcdUpdate();         
}



void Decrease(void)
{
  if((user_data[current_mode]>user_data_min[current_mode])&& current_mode==2)
    user_data[current_mode]-=10;
  else if(current_mode==3)
    user_data[current_mode]=0;       
  else if((user_data[current_mode]>user_data_min[current_mode])) 
    user_data[current_mode]-=5;
  if(user_data[current_mode]<user_data_min[current_mode])
    user_data[current_mode]=user_data_min[current_mode];
  LcdUpdate(); 
}

void start(int comp)
{  
  if(comp==0||comp==3)
  {
    if(pump_dir_value[0]==0)
      OCR2B = pwmvalues[0]; //3rd pin PWM register
    else if(pump_dir_value[0]==1)
      OCR2B = 255-pwmvalues[0]; //invert vsalues
    digitalWrite(pump_dir_pin[0],pump_dir_value[0]);
    digitalWrite(pump_enable_pin,HIGH);
  }
  else if(comp==1||comp==3)
  {
    analogWrite(pwmpins[1],pwmvalues[1]);
  }
  else if(comp==2||comp==3)
  {
    if(pump_dir_value[1]==0) 
      OCR2A = pwmvalues[2]; // 11th pin's PWM register 
    else if(pump_dir_value[1]==1)
      OCR2A = 255-pwmvalues[2];
    digitalWrite(pump_dir_pin[1],pump_dir_value[1]); 
    digitalWrite(pump_enable_pin,HIGH);
  }  

  if(comp==0)
  { 
    for(int k=0;k<4;k++)
    {
      EEPROM.write(eprom_address[k],user_data[k]);
    }
  }


}
void stopit(int comp1)
{
  if(comp1==2||comp1==3)
  {
    digitalWrite(pump_enable_pin,LOW);
    OCR2A=0;
    digitalWrite(pump_dir_pin[1],LOW);
  }
  if (comp1==1||comp1==3)
    analogWrite(pwmpins[1],0);
  if(comp1==0||comp1==3)
  { 
    OCR2B = 0; 
    digitalWrite(pump_dir_pin[0],LOW);
  }
  state=0;      
}

void calculate_parameters()
{
  float temp1=0;
  if((user_data[0]-activated_volume)<15)
    batch_volume=(user_data[0]-activated_volume);
  else
    batch_volume=15;

  //Calculate the timings
  temp1=(float)(rate1)*(float)batch_volume;
  duration_start[0]=0;  //pump1 start
  duration_end[0]=ceil(temp1);  //pump1 stop
  duration_start[1]=duration_end[0];
  duration_end[1]=duration_start[1]+user_data[2];
  duration_start[2]=duration_end[1];
  temp1=((float)rate2*(float)batch_volume);
  duration_end[2]=duration_start[2]+ceil(temp1)+5; // 5s is because pumping out last bit takes time 
  if((user_data[0])<15)
    total_duration=(ceil((float)user_data[0]*rate1)+ceil((float)user_data[0]*rate2)+5+user_data[2]);
  else
    total_duration=ceil((float)user_data[0]*rate1)+ceil((float)user_data[0]*rate2)+(6+user_data[2])*batches;

}



void timer(void)
{
  rate1=((float)1/(float)pump1_rate)*(float)60; // pumping time seconds/ml)
  rate2=((float)1/(float)pump2_rate)*(float)60; //Pumping time Seconds/ml
  activated_volume=0;
  calculate_parameters();
  //pwm variables
  pwmvalues[0]= 255;//round((pump1_rate+11.01)/0.260); //
  pwmvalues[1]=round((user_data[1]*2.55)); //percentage to pwm
  pwmvalues[2]=255;//round((pump2_rate+11.01)/0.260); 
  batches= ceil((float)user_data[0]/15); // calculate no of iterations
  total_elapsed=0;


  for(int i=0;i<batches;i++)
  {
    calculate_parameters();
    int status1[3]={
      0,0,0    };
    elapsed_time=0;
    while(elapsed_time<=duration_end[2])
    {
      gotoXY(0,5);
      LcdString("  ");
      LcdInt(total_duration-total_elapsed);
      LcdString("S left ");


      if(elapsed_time<duration_end[0]&&elapsed_time>=duration_start[0]&&status1[0]==0)
      {
        gotoXY(0,1);
        LcdString("Pumping in  ");  
        pump_dir_value[0]=0; 
        start(0);
        status1[0]=1;
      }
      else if(elapsed_time>=duration_end[0]&&status1[0]==1)
      {
        stopit(0);
        status1[0]=0;
      }

      if(elapsed_time<duration_end[1]&&elapsed_time>=duration_start[1]&&status1[1]==0)
      {
        gotoXY(0,1);
        LcdString("Irradiating");
        if(user_data[3]==1)
        {
          pump_dir_value[1]=1;
          start(2);
        }
        start(1);
        status1[1]=1;
      }
      else if(elapsed_time>=duration_end[1]&&status1[1]==1)
      {
        stopit(1);
        stopit(2);
        status1[1]=0;
        //Serial.println("inside LED-"); 
      }
      if(elapsed_time<duration_end[2]&&elapsed_time>=duration_start[2]&&status1[2]==0)
      {
        gotoXY(0,1);
        LcdString("Pumping out");
        pump_dir_value[1]=0;
        start(2);
        status1[2]=1;

      }
      else if(elapsed_time>=duration_end[2] && status1[2]==1)
      {
        stopit(2);

      } 
      delay(100);
      if(customKeypad.getKey()) {
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()) {
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()){
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()) {
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()){
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()) {
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()) {
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()){
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()){
        batches=0; 
        break;
      }
      delay(100);
      if(customKeypad.getKey()) {
        batches=0; 
        break;
      }
      elapsed_time+=1;
      total_elapsed+=1;


    }
    stopit(2);
    if(customKeypad.getKey()) {
      batches=0; 
      break;
    }
    activated_volume=activated_volume+batch_volume;
  }  

  stopit(3);
  gotoXY(0,5);
  LcdString("    Done    ");
  LcdUpdate();   

}


void setup(void)
{

  //EEPROM retrivel
  for(int k=0;k<4;k++)
  {
    user_data[k]=EEPROM.read(eprom_address[k]);
    if(user_data[k]>user_data_max[k] ||user_data[k]<user_data_min[k])
    {
      EEPROM.write(eprom_address[k],user_data[k]<user_data_min[k]);
      user_data[k]=user_data_min[k];
    }
  } 


  //LCD setup.
  LcdInitialise();
  LcdClear();        
  gotoXY(0,0);
  LcdString(" LIPs-Stick");
  gotoXY(0,5);
  LcdString("M:1 +:2 -:3");
  LcdUpdate();


  //pump intialzation
  pinMode(pump_enable_pin,OUTPUT);
  pinMode(pump_dir_pin[0],OUTPUT);
  pinMode(pump_dir_pin[1],OUTPUT);
  pinMode(pwmpins[0],OUTPUT);
  pinMode(pwmpins[1],OUTPUT);
  pinMode(pwmpins[2],OUTPUT);
  digitalWrite(pump_enable_pin,LOW);
  digitalWrite(pump_dir_pin[0],LOW);
  TCCR2A = _BV(COM2A1)|_BV(COM2B1) | _BV(WGM20); //set WGM to 001 for phase correct pwm and COM2B to 10 for non inverting PWM 
  TCCR2B = _BV(CS22); //set clock prescale value to 8
  OCR2B = 0; //pwm pin 3 3rd pin pwm value
  OCR2A=0; //pwmpin 11 register
}

void loop(){
  char customKey = customKeypad.getKey();
  if (customKey){
    if(customKey=='A')
      timer(); 
    else if(customKey=='1')
      Modechange();
    else if(customKey=='2')
      Increase();
    else if(customKey=='3')
      Decrease();
  }
}

