//Firmware to run the exposure tool in continous mode
//


#include <avr/interrupt.h>
#include <avr/io.h> 
#include <Keypad.h>
#include <EEPROM.h>
//LCD Constants 
#define PIN_RESET 7
#define PIN_SCE   8
#define PIN_DC    9
#define PIN_SDIN  10
#define PIN_SCLK  12
#define LCD_C     LOW
#define LCD_D     HIGH
#define LCD_X     84
#define LCD_Y     48
//Keypad constants
const byte ROWS = 2; //four rows
const byte COLS = 4; //four columns


//constants for converting pumping rate(ml/min) to pwm vakues. The constants can be computed by curve fitting a straight line between pwm values and pump rate.
float a_pump1=3.846,b_pump1=47.346;//pwmvalue of pump2 =a_pump1*pumping rate+b_pump1
float a_pump2=3.918,b_pump2=51.12;//pwmvalue of pump2= a_pump2*pumping rate+b_pump2
int led_delay=5,pump2_delay=10;//Time in seconds after the led and pump2 starts
int pwmvalues_min[3]={12, 50,12},pwmvalues_max[3]={54, 255 ,52};// minimum and maximum pumping rate {pump1_rate,LED_PWMvalues,pump2_rate}

//define the cymbols on the buttons of the keypads
char hexaKeys[ROWS][COLS] = {
  {'1','2','3','A'  },{'4','5','6','B'}};
int temp;
//Constants need attention
byte rowPins[ROWS] = {A0,A1}; //connect to the row pin
byte colPins[COLS] = {A5,A4,A3,A2};//connect to the column pinouts of the keypad
byte digit;
//initialize an instance of class NewKeypad
Keypad customKeypad = Keypad( makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS); 
int pwmvalues[3]={0, 0 ,0},eprom_address[11]={0,4,8,12,16,20,24,28,32,36,40};
int current_mode=0;
byte state=0;
int pwmpins[3]={3,6,11}; //3 and 11 are coupled to timer 2
int pump_dir_pin[2]={4,5};
int pump_enable=2;
int j=0,duration_start[3]={0,0,0},duration_end[3]={30,30,30},elapsed_time=0,pump_dir_value[2]={0,0};

static const byte ASCII[][5] =
{
  {
    0x00, 0x00, 0x00, 0x00, 0x00  } // 20  
  ,{
    0x00, 0x00, 0x5f, 0x00, 0x00  } // 21 !
  ,{
    0x00, 0x07, 0x00, 0x07, 0x00  } // 22 "
  ,{
    0x14, 0x7f, 0x14, 0x7f, 0x14  } // 23 #
  ,{
    0x24, 0x2a, 0x7f, 0x2a, 0x12  } // 24 $
  ,{
    0x23, 0x13, 0x08, 0x64, 0x62  } // 25 %
  ,{
    0x36, 0x49, 0x55, 0x22, 0x50  } // 26 &
  ,{
    0x00, 0x05, 0x03, 0x00, 0x00  } // 27 '
  ,{
    0x00, 0x1c, 0x22, 0x41, 0x00  } // 28 (
  ,{
    0x00, 0x41, 0x22, 0x1c, 0x00  } // 29 )
  ,{
    0x14, 0x08, 0x3e, 0x08, 0x14  } // 2a *
  ,{
    0x08, 0x08, 0x3e, 0x08, 0x08  } // 2b +
  ,{
    0x00, 0x50, 0x30, 0x00, 0x00  } // 2c ,
  ,{
    0x08, 0x08, 0x08, 0x08, 0x08  } // 2d -
  ,{
    0x00, 0x60, 0x60, 0x00, 0x00  } // 2e .
  ,{
    0x20, 0x10, 0x08, 0x04, 0x02  } // 2f /
  ,{
    0x3e, 0x51, 0x49, 0x45, 0x3e  } // 30 0
  ,{
    0x00, 0x42, 0x7f, 0x40, 0x00  } // 31 1
  ,{
    0x42, 0x61, 0x51, 0x49, 0x46  } // 32 2
  ,{
    0x21, 0x41, 0x45, 0x4b, 0x31  } // 33 3
  ,{
    0x18, 0x14, 0x12, 0x7f, 0x10  } // 34 4
  ,{
    0x27, 0x45, 0x45, 0x45, 0x39  } // 35 5
  ,{
    0x3c, 0x4a, 0x49, 0x49, 0x30  } // 36 6
  ,{
    0x01, 0x71, 0x09, 0x05, 0x03  } // 37 7
  ,{
    0x36, 0x49, 0x49, 0x49, 0x36  } // 38 8
  ,{
    0x06, 0x49, 0x49, 0x29, 0x1e  } // 39 9
  ,{
    0x00, 0x36, 0x36, 0x00, 0x00  } // 3a :
  ,{
    0x00, 0x56, 0x36, 0x00, 0x00  } // 3b ;
  ,{
    0x08, 0x14, 0x22, 0x41, 0x00  } // 3c <
  ,{
    0x14, 0x14, 0x14, 0x14, 0x14  } // 3d =
  ,{
    0x00, 0x41, 0x22, 0x14, 0x08  } // 3e >
  ,{
    0x02, 0x01, 0x51, 0x09, 0x06  } // 3f ?
  ,{
    0x32, 0x49, 0x79, 0x41, 0x3e  } // 40 @
  ,{
    0x7e, 0x11, 0x11, 0x11, 0x7e  } // 41 A
  ,{
    0x7f, 0x49, 0x49, 0x49, 0x36  } // 42 B
  ,{
    0x3e, 0x41, 0x41, 0x41, 0x22  } // 43 C
  ,{
    0x7f, 0x41, 0x41, 0x22, 0x1c  } // 44 D
  ,{
    0x7f, 0x49, 0x49, 0x49, 0x41  } // 45 E
  ,{
    0x7f, 0x09, 0x09, 0x09, 0x01  } // 46 F
  ,{
    0x3e, 0x41, 0x49, 0x49, 0x7a  } // 47 G
  ,{
    0x7f, 0x08, 0x08, 0x08, 0x7f  } // 48 H
  ,{
    0x00, 0x41, 0x7f, 0x41, 0x00  } // 49 I
  ,{
    0x20, 0x40, 0x41, 0x3f, 0x01  } // 4a J
  ,{
    0x7f, 0x08, 0x14, 0x22, 0x41  } // 4b K
  ,{
    0x7f, 0x40, 0x40, 0x40, 0x40  } // 4c L
  ,{
    0x7f, 0x02, 0x0c, 0x02, 0x7f  } // 4d M
  ,{
    0x7f, 0x04, 0x08, 0x10, 0x7f  } // 4e N
  ,{
    0x3e, 0x41, 0x41, 0x41, 0x3e  } // 4f O
  ,{
    0x7f, 0x09, 0x09, 0x09, 0x06  } // 50 P
  ,{
    0x3e, 0x41, 0x51, 0x21, 0x5e  } // 51 Q
  ,{
    0x7f, 0x09, 0x19, 0x29, 0x46  } // 52 R
  ,{
    0x46, 0x49, 0x49, 0x49, 0x31  } // 53 S
  ,{
    0x01, 0x01, 0x7f, 0x01, 0x01  } // 54 T
  ,{
    0x3f, 0x40, 0x40, 0x40, 0x3f  } // 55 U
  ,{
    0x1f, 0x20, 0x40, 0x20, 0x1f  } // 56 V
  ,{
    0x3f, 0x40, 0x38, 0x40, 0x3f  } // 57 W
  ,{
    0x63, 0x14, 0x08, 0x14, 0x63  } // 58 X
  ,{
    0x07, 0x08, 0x70, 0x08, 0x07  } // 59 Y
  ,{
    0x61, 0x51, 0x49, 0x45, 0x43  } // 5a Z
  ,{
    0x00, 0x7f, 0x41, 0x41, 0x00  } // 5b [
  ,{
    0x02, 0x04, 0x08, 0x10, 0x20  } // 5c ¥
  ,{
    0x00, 0x41, 0x41, 0x7f, 0x00  } // 5d ]
  ,{
    0x04, 0x02, 0x01, 0x02, 0x04  } // 5e ^
  ,{
    0x40, 0x40, 0x40, 0x40, 0x40  } // 5f _
  ,{
    0x00, 0x01, 0x02, 0x04, 0x00  } // 60 `
  ,{
    0x20, 0x54, 0x54, 0x54, 0x78  } // 61 a
  ,{
    0x7f, 0x48, 0x44, 0x44, 0x38  } // 62 b
  ,{
    0x38, 0x44, 0x44, 0x44, 0x20  } // 63 c
  ,{
    0x38, 0x44, 0x44, 0x48, 0x7f  } // 64 d
  ,{
    0x38, 0x54, 0x54, 0x54, 0x18  } // 65 e
  ,{
    0x08, 0x7e, 0x09, 0x01, 0x02  } // 66 f
  ,{
    0x0c, 0x52, 0x52, 0x52, 0x3e  } // 67 g
  ,{
    0x7f, 0x08, 0x04, 0x04, 0x78  } // 68 h
  ,{
    0x00, 0x44, 0x7d, 0x40, 0x00  } // 69 i
  ,{
    0x20, 0x40, 0x44, 0x3d, 0x00  } // 6a j 
  ,{
    0x7f, 0x10, 0x28, 0x44, 0x00  } // 6b k
  ,{
    0x00, 0x41, 0x7f, 0x40, 0x00  } // 6c l
  ,{
    0x7c, 0x04, 0x18, 0x04, 0x78  } // 6d m
  ,{
    0x7c, 0x08, 0x04, 0x04, 0x78  } // 6e n
  ,{
    0x38, 0x44, 0x44, 0x44, 0x38  } // 6f o
  ,{
    0x7c, 0x14, 0x14, 0x14, 0x08  } // 70 p
  ,{
    0x08, 0x14, 0x14, 0x18, 0x7c  } // 71 q
  ,{
    0x7c, 0x08, 0x04, 0x04, 0x08  } // 72 r
  ,{
    0x48, 0x54, 0x54, 0x54, 0x20  } // 73 s
  ,{
    0x04, 0x3f, 0x44, 0x40, 0x20  } // 74 t
  ,{
    0x3c, 0x40, 0x40, 0x20, 0x7c  } // 75 u
  ,{
    0x1c, 0x20, 0x40, 0x20, 0x1c  } // 76 v
  ,{
    0x3c, 0x40, 0x30, 0x40, 0x3c  } // 77 w
  ,{
    0x44, 0x28, 0x10, 0x28, 0x44  } // 78 x
  ,{
    0x0c, 0x50, 0x50, 0x50, 0x3c  } // 79 y
  ,{
    0x44, 0x64, 0x54, 0x4c, 0x44  } // 7a z
  ,{
    0x00, 0x08, 0x36, 0x41, 0x00  } // 7b {
  ,{
    0x00, 0x00, 0x7f, 0x00, 0x00  } // 7c |
  ,{
    0x00, 0x41, 0x36, 0x08, 0x00  } // 7d }
  ,{
    0x10, 0x08, 0x08, 0x10, 0x08  } // 7e ←
  ,{
    0x78, 0x46, 0x41, 0x46, 0x78  } // 7f →
};

void LcdCharacter(char character)
{
  LcdWrite(LCD_D, 0x00);
  for (int index = 0; index < 5; index++)
  {
    LcdWrite(LCD_D, ASCII[character - 0x20][index]);
  }
  LcdWrite(LCD_D, 0x00);
}

void LcdClear(void)
{
  for (int index = 0; index < LCD_X * LCD_Y / 8; index++)
  {
    LcdWrite(LCD_D, 0x00);
  }
}

void LcdInitialise(void)
{
  pinMode(PIN_SCE, OUTPUT);
  pinMode(PIN_RESET, OUTPUT);
  pinMode(PIN_DC, OUTPUT);
  pinMode(PIN_SDIN, OUTPUT);
  pinMode(PIN_SCLK, OUTPUT);
  digitalWrite(PIN_RESET, LOW);
  digitalWrite(PIN_RESET, HIGH);
  LcdWrite(LCD_C, 0x21 );  // LCD Extended Commands.
  LcdWrite(LCD_C, 0xB1 );  // Set LCD Vop (Contrast). 
  LcdWrite(LCD_C, 0x04 );  // Set Temp coefficent. //0x04
  LcdWrite(LCD_C, 0x14 );  // LCD bias mode 1:48. //0x13
  LcdWrite(LCD_C, 0x20 );  // LCD Basic Commands
  LcdWrite(LCD_C, 0x0C );  // LCD in normal mode.
}

void LcdString(char *characters)
{
  while (*characters)
  {
    LcdCharacter(*characters++);
  }
}

void LcdWrite(byte dc, byte data)
{
  digitalWrite(PIN_DC, dc);
  digitalWrite(PIN_SCE, LOW);
  shiftOut(PIN_SDIN, PIN_SCLK, MSBFIRST, data);
  digitalWrite(PIN_SCE, HIGH);
}
void gotoXY(int x, int y)
{
  LcdWrite( 0, 0x80 | x);  // Column.
  LcdWrite( 0, 0x40 | y);  // Row.  

}
void LcdInt(int number)
{
  int temp_var=number;
  int count=0;
  byte digit[3]={
    0,0,0  };
  do
  {
    digit[count]=temp_var%10;
    temp_var=temp_var/10;
    count++;
  } 
  while(temp_var!=0);
  if(digit[2]!=0)
  {
    LcdWrite(LCD_D, 0x00);
    for (int index = 0; index < 5; index++)
    {
      LcdWrite(LCD_D, ASCII[digit[2]+0x10][index]);
    }
    LcdWrite(LCD_D, 0x00);
  }

  for(j=1;j>=0;j--)
  {
    LcdWrite(LCD_D, 0x00);
    for (int index = 0; index < 5; index++)
    {
      LcdWrite(LCD_D, ASCII[digit[j]+0x10][index]);
    }
    LcdWrite(LCD_D, 0x00);
  }  
}
void LcdUpdate()
{
  //Mode Name
  gotoXY(0,1);
  if(current_mode==0)
    LcdString("Mode:Pump1  ");
  else if(current_mode==1)
    LcdString("Mode:LED   ");
  else if(current_mode==2)
    LcdString("Mode:Pump2   ");
  else if(current_mode==3)
    LcdString("Mode:P1_DIR  ");
  else if(current_mode==4)
    LcdString("Mode:P2_DIR  ");
  else if(current_mode==5)
    LcdString("Mode:Time     ");

  //Values update.
  gotoXY(0,2);
  LcdString("Pump1:");
  if(pump_dir_value[0]==HIGH)
    LcdString("+");
  else if(pump_dir_value[0]==LOW)
    LcdString("-");
  LcdInt(pwmvalues[0]); 
  //  LcdString("/");
  //  LcdInt(duration_end[0]);

  gotoXY(0,3);
  LcdString("LED:");
  LcdInt(pwmvalues[1]);
  LcdString("/");
  LcdInt(duration_end[1]);


  gotoXY(0,4);
  LcdString("pump2:");
  if(pump_dir_value[1]==HIGH)
    LcdString("+");
  else if(pump_dir_value[1]==LOW)
    LcdString("-");
  LcdInt(pwmvalues[2]);

}




//KEYPAD functions
void Modechange(void)
{
  if(current_mode<5)
    current_mode++;
  else if(current_mode>=5)
    current_mode=0;
  LcdUpdate();  
}


void Increase(void)
{
  int temp1=0,temp2=0,temp3=0;
  if((pwmvalues[current_mode]<pwmvalues_max[current_mode])&&(pwmvalues[current_mode]>=pwmvalues_min[current_mode])&& current_mode==1)
    pwmvalues[current_mode]+=5;
  else if((pwmvalues[current_mode]<pwmvalues_max[current_mode])&&(pwmvalues[current_mode]>=pwmvalues_min[current_mode])) 
    pwmvalues[current_mode]+=1;
  else if(pwmvalues[current_mode]<pwmvalues_min[current_mode])
    pwmvalues[current_mode]=pwmvalues_min[current_mode];

  if(current_mode==3||current_mode==4)
  {
    temp1=current_mode-3;
    pump_dir_value[temp1]=1;
    current_mode=temp1+3;

  }

  else if(current_mode==5)
  {

    duration_end[1]+=5;


    //         pump_dir_value[0]=temp2;
    //         pump_dir_value[1]=temp3;
    //         duration_start[1]=0;
    //         duration_start[2]=0;
  }

  LcdUpdate();



}
void Decrease(void)
{
  int temp1=0,temp2=0,temp3=0;
  if(pwmvalues[current_mode]>pwmvalues_min[current_mode]&&current_mode==1)
    pwmvalues[current_mode]-=5;
  else if(pwmvalues[current_mode]>pwmvalues_min[current_mode])
    pwmvalues[current_mode]-=1;
  else if(pwmvalues[current_mode]<=pwmvalues_min[current_mode])
    pwmvalues[current_mode]=0;

  if(current_mode==3||current_mode==4)
  {
    temp1=current_mode-3;
    pump_dir_value[temp1]=0;
    current_mode=temp1+3;

  }

  else if(current_mode>=5&&duration_end[1]>=5) 
    duration_end[1]-=5;
  LcdUpdate();
}
void start(int comp)
{

  if(comp==0||comp==3)
  {
    if(pump_dir_value[0]==0)
      OCR2B = round(a_pump1*pwmvalues[0]+b_pump1); //3rd pin 11th pin's register OCR2A
    else if(pump_dir_value[0]==1)
      OCR2B = 255- round(a_pump1*pwmvalues[0]+b_pump1); 
    digitalWrite(pump_dir_pin[0],pump_dir_value[0]);
    digitalWrite(pump_enable,HIGH);
  }
  else if(comp==1||comp==3)
  {
    analogWrite(pwmpins[1],pwmvalues[1]);
  }
  else if(comp==2||comp==3)
  {
    if(pump_dir_value[1]==0)//11th pin 
      OCR2A = round((a_pump2*pwmvalues[2]+b_pump2)); //3rd pin 11th pin's register OCR2A
    else if(pump_dir_value[1]==1)
      OCR2A = 255- round((a_pump2*pwmvalues[2]+b_pump2));
    digitalWrite(pump_dir_pin[1],pump_dir_value[1]); 
  }  

  if(comp==0)
  {
    //state=1; 
    for(int k=0;k<3;k++)
    {
      EEPROM.write(eprom_address[k],pwmvalues[k]);
    }
    EEPROM.write(eprom_address[3],pump_dir_value[0]);
    EEPROM.write(eprom_address[4],pump_dir_value[1]);
    EEPROM.write(eprom_address[9],duration_end[9]);
  }


}
void stopit(int comp)
{
  if(comp==2||comp==3)
  {
    digitalWrite(pump_enable,LOW);
    OCR2A=0;
    digitalWrite(pump_dir_pin[1],LOW);
  }
  if (comp==1||comp==3)
    analogWrite(pwmpins[1],0);
  else if(comp==0||comp==3)
  { 
    OCR2B = 0; 
    digitalWrite(pump_dir_pin[0],LOW);
  }
  state=0;      
}
void timer(void)
{
  //setting up duration _start
  duration_start[1]=led_delay; //When to start LED
  duration_start[2]=pump2_delay; //When to start motor
  int status1[3]={0,0,0};
  elapsed_time=0;
  int total_duration=duration_start[1]+duration_end[1]+duration_start[2];
  while(elapsed_time<=(total_duration))
  {
    gotoXY(0,5);
    LcdString("   ");
    LcdInt((total_duration-elapsed_time));
    LcdString("S left");

    if(status1[0]==0)
    {
      start(0);
      status1[0]=1;
    }
    else if(elapsed_time>=(total_duration-duration_start[1])&&status1[0]==1)
    {
      stopit(0); 
    }

    if(elapsed_time>duration_start[1]&&status1[1]==0)
    {
      start(1);
      status1[1]=1;
    }
    else if(elapsed_time>=total_duration&&status1[1]==1)
    {
      stopit(1);
      status1[1]=0;
    }
    if(elapsed_time>duration_start[2]&&status1[2]==0)
    {
      start(2);
      status1[2]=1;
    }
    else if(elapsed_time>=total_duration&& status1[2]==1)
    {
      stopit(2);
    } 
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    delay(100);
    if(customKeypad.getKey()) break;
    elapsed_time+=1;


  }
  stopit(3);
  gotoXY(0,5);
  LcdString("  Stopped   ");   
}


void setup(void)
{
  pinMode(pwmpins[2],OUTPUT);
  analogWrite(pwmpins[2],0);
  //EEPROM retrivel
  for(int k=0;k<3;k++)
  {
    pwmvalues[k]=EEPROM.read(eprom_address[k]);
  }
  for(int k=0;k<2;k++)
  {
    pump_dir_value[k]= EEPROM.read(eprom_address[(k+3)]);
    if(pump_dir_value[k]>1)  //the eeprom will 255 initially
    {
      EEPROM.write(eprom_address[(k+3)],0);
      pump_dir_value[k]=0;
    }
  }
  duration_end[1]= EEPROM.read(eprom_address[9]);
  if(duration_end[1]<5)  //the eeprom will 255 initially
  {

    EEPROM.write(eprom_address[9],30);
    duration_end[1]=30;
  }
  //LCD setup.
  LcdInitialise();
  LcdClear();        
  gotoXY(0,0);
  LcdString(" LIPs-Stick");
  gotoXY(0,5);
  LcdString("M:1 +:2 -:3");
  LcdUpdate();

  //pump 1&2
  pinMode(pump_enable,OUTPUT);
  pinMode(pump_dir_pin[0],OUTPUT);
  pinMode(pump_dir_pin[1],OUTPUT);
  pinMode(pwmpins[0],OUTPUT);
  pinMode(pwmpins[1],OUTPUT);
  digitalWrite(pump_enable,LOW);
  digitalWrite(pump_dir_pin[0],LOW);
  TCCR2A = _BV(COM2A1)|_BV(COM2B1) | _BV(WGM20); //set WGM to 001 for phase correct pwm and COM2B to 10 for non inverting PWM 
  TCCR2B = _BV(CS22); //set clock prescale value to 8
  OCR2B = 0; //pwm pin[0] - 3rd pin pwm value
  OCR2A=0; //pwmpin 11 register
  Serial.begin(9600);
}

void loop(){
  char customKey = customKeypad.getKey();

  if (customKey){
    Serial.println(customKey);

    if(customKey=='A')
      timer(); 
    else if(customKey=='1')
      Modechange();
    else if(customKey=='2')
      Increase();
    else if(customKey=='3')
      Decrease();

  }
}

